# Generated by Django 4.2.18 on 2025-01-20 13:24

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pfxcore', '0002_pfxpermissionsuser'),
    ]

    operations = [
        migrations.DeleteModel(
            name='PFXPermissionsUser',
        ),
    ]
