# Generated by Django 4.2.17 on 2025-01-21 09:43
# flake8: noqa

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]
