# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import django
from django.conf import settings

from sphinx.ext.autodoc import between

sys.path.insert(0, os.path.abspath('../'))
settings.configure(
    SECRET_KEY='something to make Django happy',
    INSTALLED_APPS=[
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.postgres',
        'pfx.pfxcore'])
django.setup()


def setup(app):
    # Register a sphinx.ext.autodoc.between listener to ignore everything
    # between lines that contain --- and """ (yaml from apispec)
    app.connect('autodoc-process-docstring',
                between('^.*(---|""")$', exclude=True))
    return app


# -- Project information -----------------------------------------------------
project = 'Django PFX'
copyright = '2021, Hervé Martinet, Nicolas Béraneck'
author = 'Hervé Martinet, Nicolas Béraneck'

# The full version, including alpha/beta/rc tags
release = '1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['myst_parser',
              'sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinxcontrib.mermaid']
autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

myst_heading_anchors = 3
